"""
MAUS s (plural of MAUS) are data structures which are structurally equivalent to the maus.DeepAnwendungshandbuch.
For more information see the MAUS repo and its documentation: https://github.com/Hochfrequenz/mig_ahb_utility_stack/
A MAUS Provider is a class that returns MAUS s' from what ever data source the implementation prefers.
The MAUS provider is supposed to be used with dependency injection.
"""

import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

from maus import DeepAnwendungshandbuch
from maus.edifact import EdifactFormat, EdifactFormatVersion
from maus.models.anwendungshandbuch import DeepAnwendungshandbuchSchema

# pylint:disable=too-few-public-methods


class MausProvider(ABC):
    """
    A MausProvider is a class that provides MAUS' (Deep Anwendungshandbuch) to calling code.
    """

    @abstractmethod
    def get_maus(
        self, edifact_format: EdifactFormat, edifact_format_version: EdifactFormatVersion, pruefidentifikator: str
    ) -> Optional[DeepAnwendungshandbuch]:
        """
        Return a MAUS for the given parameters. returns None if the requested MAUS is not available.
        """
        raise NotImplementedError("Has to be implemented in inheriting class")


class FileBasedMausProvider(MausProvider):
    """
    A MAUS provider that uses the file system to retrieve MAUS s.
    """

    def __init__(self, base_path: Path, encoding: str = "utf-8"):
        """
        initialize by providing a base path relative to which the MAUS s can be found.
        """
        self.base_path: Path = base_path
        self._encoding = encoding

    @abstractmethod
    def to_path(
        self, edifact_format: EdifactFormat, edifact_format_version: EdifactFormatVersion, pruefidentifikator: str
    ) -> Path:
        """
        returns the path of the maus file relative to the given parameters.
        """
        raise NotImplementedError("Has to be implemented in inheriting class")

    def get_maus(
        self, edifact_format: EdifactFormat, edifact_format_version: EdifactFormatVersion, pruefidentifikator: str
    ) -> Optional[DeepAnwendungshandbuch]:
        relative_path = self.to_path(edifact_format, edifact_format_version, pruefidentifikator)
        full_path: Path = self.base_path / relative_path
        try:
            with open(full_path, "r", encoding=self._encoding) as maus_infile:
                file_content_json = json.load(maus_infile)
                maus = DeepAnwendungshandbuchSchema().load(file_content_json)
        except FileNotFoundError:
            return None
        return maus



"""
A module to generate ready to use evaluators.

AHBicht, in the first place, is designed to do the content evaluation itself with actual code that checks if conditions
are fulfilled, format constraints are met and gets hints from somewhere. However, AHBicht can also be used to do a
"fake" content evaluation where the single truth values of conditions are already known.
Then the approach with implementing single methods whose outcomes are already known/hardcoded at compile time
seems like a big overhead. The code representation of "all outcomes are already known" is an instance of the
ContentEvaluationResult. Now the methods below are useful. Simply provide a content evaluation result (the data) and
the evaluators are created based on the already known outcomes. You do not have to actually touch any evaluator code.
"""
from typing import Tuple

import inject

from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResult
from ahbicht.content_evaluation.fc_evaluators import DictBasedFcEvaluator, FcEvaluator
from ahbicht.content_evaluation.rc_evaluators import DictBasedRcEvaluator, RcEvaluator
from ahbicht.expressions.hints_provider import DictBasedHintsProvider, HintsProvider
from ahbicht.expressions.package_expansion import DictBasedPackageResolver, PackageResolver


def create_hardcoded_evaluators(
    content_evaluation_result: ContentEvaluationResult,
) -> Tuple[RcEvaluator, FcEvaluator, HintsProvider, PackageResolver]:
    """
    Creates evaluators based on the given content_evaluation_result

    :param content_evaluation_result:
    :return:
    """
    rc_evaluator = DictBasedRcEvaluator(content_evaluation_result.requirement_constraints)
    fc_evaluator = DictBasedFcEvaluator(content_evaluation_result.format_constraints)
    hints_provider = DictBasedHintsProvider(content_evaluation_result.hints)
    package_resolver = DictBasedPackageResolver(content_evaluation_result.packages or {})
    return rc_evaluator, fc_evaluator, hints_provider, package_resolver


def create_and_inject_hardcoded_evaluators(content_evaluation_result: ContentEvaluationResult):
    """
    Creates evaluators from hardcoded content_evaluation result and injects them

    :param content_evaluation_result:
    :return:
    """
    evaluators = create_hardcoded_evaluators(content_evaluation_result)
    inject.clear_and_configure(
        lambda binder: binder.bind(RcEvaluator, evaluators[0])  # type:ignore[arg-type]
        .bind(FcEvaluator, evaluators[1])
        .bind(HintsProvider, evaluators[2])
        .bind(PackageResolver, evaluators[3])
    )



"This module contains the classes for the validation results."

from abc import ABC
from typing import Dict, Optional

import attrs
from marshmallow import Schema, fields, post_load
from marshmallow_enum import EnumField  # type:ignore[import]

from ahbicht.validation.validation_values import RequirementValidationValue

# pylint: disable=too-few-public-methods, no-member, no-self-use, unused-argument


@attrs.define(auto_attribs=True, kw_only=True)
class ValidationResult(ABC):
    """Result of the validation"""

    #: In which requirement state is the field and is it filled or not?
    requirement_validation: RequirementValidationValue = attrs.field(
        validator=attrs.validators.instance_of(RequirementValidationValue)
    )
    #: Collected hints
    hints: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)), default=None
    )


class ValidationResultSchema(Schema):
    """
    A schema to (de-)serialize ValidationResult
    """

    requirement_validation = EnumField(RequirementValidationValue)
    hints = fields.String(load_default=None)

    @post_load
    def deserialize(self, data, **kwargs) -> ValidationResult:
        """
        Converts the barely typed data dictionary into an actual ValidationResult
        :param data:
        :param kwargs:
        :return:
        """
        return ValidationResult(**data)


@attrs.define(auto_attribs=True, kw_only=True)
class SegmentLevelValidationResult(ValidationResult):
    """Result of the validation of a segment or segment group"""


class SegmentLevelValidationResultSchema(ValidationResultSchema):
    """
    A schema to (de-)serialize SegmentLevelValidationResult
    """

    @post_load
    def deserialize(self, data, **kwargs) -> SegmentLevelValidationResult:
        """
        Converts the barely typed data dictionary into an actual ValidationResult
        :param data:
        :param kwargs:
        :return:
        """
        return SegmentLevelValidationResult(**data)


@attrs.define(auto_attribs=True, kw_only=True)
class DataElementValidationResult(ValidationResult):
    """Result of the validation of a data element"""

    #: Is the format constraint fulfilled or not?
    format_validation_fulfilled: bool = attrs.field(validator=attrs.validators.instance_of(bool))
    #: possible error message regarding the format constraints
    format_error_message: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)), default=None
    )
    #: possible qualifiers for value pool data elements
    possible_values: Optional[Dict[str, str]] = attrs.field(
        default=None,
        validator=attrs.validators.optional(
            attrs.validators.deep_mapping(
                key_validator=attrs.validators.instance_of(str),
                value_validator=attrs.validators.instance_of(str),
                mapping_validator=attrs.validators.instance_of(dict),
            )
        ),
    )


class DataElementValidationResultSchema(ValidationResultSchema):
    """
    A schema to (de-)serialize DataElementValidationResult
    """

    format_validation_fulfilled = fields.Bool()
    format_error_message = fields.String(load_default=None)
    possible_values = fields.Dict(fields.Str, load_default=None)

    @post_load
    def deserialize(self, data, **kwargs) -> DataElementValidationResult:
        """
        Converts the barely typed data dictionary into an actual DataElementValidationResult
        :param data:
        :param kwargs:
        :return:
        """
        return DataElementValidationResult(**data)


@attrs.define(auto_attribs=True, kw_only=True)
class ValidationResultInContext:
    """
    Class to set validation result in context, for example with its discriminator.
    """

    discriminator: str = attrs.field(validator=attrs.validators.instance_of(str))
    validation_result: ValidationResult = attrs.field(validator=attrs.validators.instance_of(ValidationResult))


class ValidationResultInContextSchema(Schema):
    """
    A schema to serialize ValidationResultInContext
    """

    discriminator = fields.String()
    validation_result = fields.Nested(ValidationResultSchema)

    # pylint:disable=unused-argument,no-self-use
    @post_load
    def deserialize(self, data, **kwargs) -> ValidationResultInContext:
        """
        Serializes a ValidationResultDiscriminator
        so that the discriminator is the key and the validation_result the value.
        """
        return ValidationResultInContext(**data)



# type:ignore
# MyPy will (reproducible) run in an infinite loop if you remove the type ignore in the test! We literally killed it 💀
"""
Tests that the parsed trees are JSON serializable
"""

import pytest  # type:ignore[import]
from lark import Token, Tree

from ahbicht.expressions.ahb_expression_parser import parse_ahb_expression_to_single_requirement_indicator_expressions
from ahbicht.expressions.condition_expression_parser import parse_condition_expression_to_tree
from ahbicht.expressions.expression_resolver import parse_expression_including_unresolved_subexpressions
from ahbicht.json_serialization.concise_condition_key_tree_schema import ConciseConditionKeyTreeSchema
from ahbicht.json_serialization.concise_tree_schema import ConciseTreeSchema
from ahbicht.json_serialization.tree_schema import TreeSchema
from unittests.test_json_serialization import _test_serialization_roundtrip  # type:ignore[import]

pytestmark = pytest.mark.asyncio


class TestTreeSchemas:
    """
    This class tests the serialization of trees (because there are multiple schemas for it).
    All the other (non tree) serialization tests happen in :class:`TestJsonSerialization`.
    """

    @pytest.mark.parametrize(
        "tree, expected_json_dict",
        [
            pytest.param(
                Tree(
                    "or_composition",
                    [
                        Tree("condition_key", [Token("INT", "53")]),
                        Tree(
                            "and_composition",
                            [Tree("condition_key", [Token("INT", "1")]), Tree("condition_key", [Token("INT", "2")])],
                        ),
                    ],
                ),
                {
                    "children": [
                        {
                            "token": None,
                            "tree": {
                                "children": [{"token": {"type": "INT", "value": "53"}, "tree": None}],
                                "type": "condition_key",
                            },
                        },
                        {
                            "token": None,
                            "tree": {
                                "children": [
                                    {
                                        "token": None,
                                        "tree": {
                                            "children": [{"token": {"type": "INT", "value": "1"}, "tree": None}],
                                            "type": "condition_key",
                                        },
                                    },
                                    {
                                        "token": None,
                                        "tree": {
                                            "children": [{"token": {"type": "INT", "value": "2"}, "tree": None}],
                                            "type": "condition_key",
                                        },
                                    },
                                ],
                                "type": "and_composition",
                            },
                        },
                    ],
                    "type": "or_composition",
                },
            )
        ],
    )
    def test_tree_serialization(self, tree: Tree, expected_json_dict: dict):
        _test_serialization_roundtrip(tree, TreeSchema(), expected_json_dict)

    @pytest.mark.parametrize(
        "condition_string, expected_json_dict",
        [
            pytest.param(
                "Muss [2] U ([3] O [4])[901] U [555]",
                {
                    "children": [
                        {
                            "token": None,
                            "tree": {
                                "children": [
                                    {"token": {"type": "MODAL_MARK", "value": "Muss"}, "tree": None},
                                    {
                                        "token": {
                                            "type": "CONDITION_EXPRESSION",
                                            "value": " [2] U ([3] O [4])[901] U [555]",
                                        },
                                        "tree": None,
                                    },
                                ],
                                "type": "single_requirement_indicator_expression",
                            },
                        }
                    ],
                    "type": "ahb_expression",
                },
            )
        ],
    )
    def test_single_requirement_indicator_expression_serialization(
        self, condition_string: str, expected_json_dict: dict
    ):
        tree = parse_ahb_expression_to_single_requirement_indicator_expressions(condition_string)
        _test_serialization_roundtrip(tree, TreeSchema(), expected_json_dict)

    @pytest.mark.parametrize(
        "condition_expression, expected_compact_json_dict",
        [
            pytest.param(
                "[1] U ([2] O [3])[901]",
                {"and_composition": ["1", {"then_also_composition": [{"or_composition": ["2", "3"]}, "901"]}]},
            ),
            pytest.param(
                "[3] U ([2] O [3] U [77] X [99][502])[901]",
                {
                    "and_composition": [
                        "3",
                        {
                            "then_also_composition": [
                                {
                                    "or_composition": [
                                        "2",
                                        {
                                            "xor_composition": [
                                                {"and_composition": ["3", "77"]},
                                                {"then_also_composition": ["99", "502"]},
                                            ]
                                        },
                                    ]
                                },
                                "901",
                            ]
                        },
                    ]
                },
            ),
        ],
    )
    def test_concise_tree_serialization_behaviour_for_condition_expressions(
        self, condition_expression: str, expected_compact_json_dict: dict
    ):
        tree = parse_condition_expression_to_tree(condition_expression)
        json_dict = ConciseTreeSchema().dump(tree)
        assert json_dict == expected_compact_json_dict

    @pytest.mark.parametrize(
        "ahb_expression, expected_compact_json_dict",
        [
            pytest.param(
                "Muss [1] U ([2] O [3])[901]",
                {
                    "ahb_expression": [
                        {
                            "single_requirement_indicator_expression": [
                                "Muss",
                                {
                                    "and_composition": [
                                        "1",
                                        {"then_also_composition": [{"or_composition": ["2", "3"]}, "901"]},
                                    ]
                                },
                            ]
                        }
                    ]
                },
            ),
            pytest.param(
                "Soll [3] U ([2] O [3] U [77] X [99][502])[901] Kann [43]",
                {
                    "ahb_expression": [
                        {
                            "single_requirement_indicator_expression": [
                                "Soll",
                                {
                                    "and_composition": [
                                        "3",
                                        {
                                            "then_also_composition": [
                                                {
                                                    "or_composition": [
                                                        "2",
                                                        {
                                                            "xor_composition": [
                                                                {"and_composition": ["3", "77"]},
                                                                {"then_also_composition": ["99", "502"]},
                                                            ]
                                                        },
                                                    ]
                                                },
                                                "901",
                                            ]
                                        },
                                    ]
                                },
                            ]
                        },
                        {"single_requirement_indicator_expression": ["Kann", "43"]},
                    ]
                },
            ),
        ],
    )
    async def test_concise_tree_serialization_behaviour_for_ahb_expressions(
        self, ahb_expression: str, expected_compact_json_dict: dict
    ):
        tree = await parse_expression_including_unresolved_subexpressions(ahb_expression)
        json_dict = ConciseTreeSchema().dump(tree)
        assert json_dict == expected_compact_json_dict

    @pytest.mark.parametrize(
        "expression, expected_compact_json_dict",
        [
            pytest.param(
                "[53] O ([1] U [2])",
                {
                    "children": [
                        {
                            "token": {
                                "value": "53",
                                "type": "condition_key",
                            },
                            "tree": None,
                        },
                        {
                            "token": None,
                            "tree": {
                                "children": [
                                    {
                                        "token": {
                                            "value": "1",
                                            "type": "condition_key",
                                        },
                                        "tree": None,
                                    },
                                    {
                                        "token": {
                                            "value": "2",
                                            "type": "condition_key",
                                        },
                                        "tree": None,
                                    },
                                ],
                                "type": "and_composition",
                            },
                        },
                    ],
                    "type": "or_composition",
                },
            )
        ],
    )
    def test_concise_condition_key_tree_serialization_behaviour_for_condition_expressions(
        self, expression: str, expected_compact_json_dict: dict
    ):
        tree = parse_condition_expression_to_tree(expression)
        json_dict = ConciseConditionKeyTreeSchema().dump(tree)
        assert json_dict == expected_compact_json_dict

    @pytest.mark.parametrize(
        "expression, expected_compact_json_dict",
        [
            pytest.param(
                "X [1]",
                {
                    "children": [
                        {
                            "token": None,
                            "tree": {
                                "children": [{"token": {"value": "1", "type": "condition_key"}, "tree": None}],
                                "type": "X",
                            },
                        }
                    ],
                    "type": "ahb_expression",
                },
            ),
            pytest.param(
                "Muss [1]",
                {
                    "children": [
                        {
                            "token": None,
                            "tree": {
                                "children": [{"token": {"value": "1", "type": "condition_key"}, "tree": None}],
                                "type": "Muss",
                            },
                        }
                    ],
                    "type": "ahb_expression",
                },
            ),
            pytest.param(
                "Muss [1] U ([2] O [3])[901]",
                {
                    "type": "ahb_expression",
                    "children": [
                        {
                            "token": None,
                            "tree": {
                                "type": "Muss",
                                "children": [
                                    {
                                        "token": None,
                                        "tree": {
                                            "type": "and_composition",
                                            "children": [
                                                {"token": {"value": "1", "type": "condition_key"}, "tree": None},
                                                {
                                                    "token": None,
                                                    "tree": {
                                                        "type": "then_also_composition",
                                                        "children": [
                                                            {
                                                                "token": None,
                                                                "tree": {
                                                                    "type": "or_composition",
                                                                    "children": [
                                                                        {
                                                                            "token": {
                                                                                "value": "2",
                                                                                "type": "condition_key",
                                                                            },
                                                                            "tree": None,
                                                                        },
                                                                        {
                                                                            "token": {
                                                                                "value": "3",
                                                                                "type": "condition_key",
                                                                            },
                                                                            "tree": None,
                                                                        },
                                                                    ],
                                                                },
                                                            },
                                                            {
                                                                "token": {"value": "901", "type": "condition_key"},
                                                                "tree": None,
                                                            },
                                                        ],
                                                    },
                                                },
                                            ],
                                        },
                                    },
                                ],
                            },
                        }
                    ],
                },
            ),
            pytest.param(
                "Soll [3] U ([2] O [3] U [77] X [99][502])[901] Kann [43]",
                {
                    "children": [
                        {
                            "token": None,
                            "tree": {
                                "children": [
                                    {
                                        "token": None,
                                        "tree": {
                                            "children": [
                                                {"token": {"value": "3", "type": "condition_key"}, "tree": None},
                                                {
                                                    "token": None,
                                                    "tree": {
                                                        "children": [
                                                            {
                                                                "token": None,
                                                                "tree": {
                                                                    "children": [
                                                                        {
                                                                            "token": {
                                                                                "value": "2",
                                                                                "type": "condition_key",
                                                                            },
                                                                            "tree": None,
                                                                        },
                                                                        {
                                                                            "token": None,
                                                                            "tree": {
                                                                                "children": [
                                                                                    {
                                                                                        "token": None,
                                                                                        "tree": {
                                                                                            "children": [
                                                                                                {
                                                                                                    "token": {
                                                                                                        "value": "3",
                                                                                                        "type": "condition_key",
                                                                                                    },
                                                                                                    "tree": None,
                                                                                                },
                                                                                                {
                                                                                                    "token": {
                                                                                                        "value": "77",
                                                                                                        "type": "condition_key",
                                                                                                    },
                                                                                                    "tree": None,
                                                                                                },
                                                                                            ],
                                                                                            "type": "and_composition",
                                                                                        },
                                                                                    },
                                                                                    {
                                                                                        "token": None,
                                                                                        "tree": {
                                                                                            "children": [
                                                                                                {
                                                                                                    "token": {
                                                                                                        "value": "99",
                                                                                                        "type": "condition_key",
                                                                                                    },
                                                                                                    "tree": None,
                                                                                                },
                                                                                                {
                                                                                                    "token": {
                                                                                                        "value": "502",
                                                                                                        "type": "condition_key",
                                                                                                    },
                                                                                                    "tree": None,
                                                                                                },
                                                                                            ],
                                                                                            "type": "then_also_composition",
                                                                                        },
                                                                                    },
                                                                                ],
                                                                                "type": "xor_composition",
                                                                            },
                                                                        },
                                                                    ],
                                                                    "type": "or_composition",
                                                                },
                                                            },
                                                            {
                                                                "token": {"value": "901", "type": "condition_key"},
                                                                "tree": None,
                                                            },
                                                        ],
                                                        "type": "then_also_composition",
                                                    },
                                                },
                                            ],
                                            "type": "and_composition",
                                        },
                                    },
                                ],
                                "type": "Soll",
                            },
                        },
                        {
                            "token": None,
                            "tree": {
                                "children": [{"token": {"value": "43", "type": "condition_key"}, "tree": None}],
                                "type": "Kann",
                            },
                        },
                    ],
                    "type": "ahb_expression",
                },
            ),
        ],
    )
    async def test_concise_condition_key_tree_serialization_behaviour_for_ahb_expressions(
        self, expression: str, expected_compact_json_dict: dict
    ):
        tree = await parse_expression_including_unresolved_subexpressions(expression)
        json_dict = ConciseConditionKeyTreeSchema().dump(tree)
        assert json_dict == expected_compact_json_dict



"""
This module evaluates the parsed ahb expression tree.
The AhbExpressionTransformer defines the rules how the different parts of the parsed tree are handled.

The used terms are defined in the README.md.
"""
from typing import Awaitable, Dict, List, Optional, Union

from lark import Token, Transformer, Tree, v_args
from lark.exceptions import VisitError

from ahbicht.evaluation_results import (
    AhbExpressionEvaluationResult,
    FormatConstraintEvaluationResult,
    RequirementConstraintEvaluationResult,
)
from ahbicht.expressions.enums import ModalMark, PrefixOperator, RequirementIndicator
from ahbicht.expressions.format_constraint_expression_evaluation import format_constraint_evaluation
from ahbicht.expressions.requirement_constraint_expression_evaluation import requirement_constraint_evaluation
from ahbicht.utility_functions import gather_if_necessary

_str_to_modal_mark_mapping: Dict[str, ModalMark] = {
    "MUSS": ModalMark.MUSS,
    "M": ModalMark.MUSS,
    "KANN": ModalMark.KANN,
    "K": ModalMark.KANN,
    "SOLL": ModalMark.SOLL,
    "S": ModalMark.SOLL,
}


# pylint: disable=no-self-use, invalid-name
# no-self-use: The following methods are not static because it refers to the terminals of the lark grammar.
# invalid-name: That's also the reason why they seemingly violate the naming conventions.
class AhbExpressionTransformer(Transformer):
    """
    Transformer, that evaluates the trees built from the ahb expressions.
    The input are the conditions as defined in the AHBs in the form of ConditionNodes.
    It returns a list of the seperated requirement indicators with
    their respective condition expressions already evaluated to booleans.
    """

    def __init__(self, entered_input: Optional[str]):
        """
        The input are the evaluated format constraint conditions in the form of ConditionNodes.

        :param entered_input: dict(condition_keys, ConditionNode)
        """
        super().__init__()
        self.entered_input = entered_input

    def CONDITION_EXPRESSION(self, condition_expression: Token) -> str:
        """Returns the condition expression."""
        return condition_expression.value

    def PREFIX_OPERATOR(self, prefix_operator: Token) -> PrefixOperator:
        """Returns the prefix operator."""
        return PrefixOperator(prefix_operator.value)

    def MODAL_MARK(self, modal_mark: Token) -> ModalMark:
        """Returns the modal mark."""
        return _str_to_modal_mark_mapping[modal_mark.value.upper()]

    @v_args(inline=True)  # Children are provided as *args instead of a list argument
    def single_requirement_indicator_expression(
        self, requirement_indicator: RequirementIndicator, condition_expression
    ) -> Awaitable[AhbExpressionEvaluationResult]:
        """
        Evaluates the condition expression of the respective requirement indicator expression and returns a list of the
        seperated requirement indicators with their results of the condition check.
        """
        return self._single_requirement_indicator_expression_async(requirement_indicator, condition_expression)

    async def _single_requirement_indicator_expression_async(
        self, requirement_indicator: RequirementIndicator, condition_expression
    ) -> AhbExpressionEvaluationResult:
        """
        See :meth:`single_requirement_indicator_expression_async`
        """
        requirement_constraint_evaluation_result: RequirementConstraintEvaluationResult = (
            await requirement_constraint_evaluation(condition_expression)
        )

        format_constraint_evaluation_result: FormatConstraintEvaluationResult = await format_constraint_evaluation(
            requirement_constraint_evaluation_result.format_constraints_expression, self.entered_input
        )

        result_of_ahb_expression_evaluation: AhbExpressionEvaluationResult = AhbExpressionEvaluationResult(
            requirement_indicator=requirement_indicator,
            requirement_constraint_evaluation_result=requirement_constraint_evaluation_result,
            format_constraint_evaluation_result=format_constraint_evaluation_result,
        )

        return result_of_ahb_expression_evaluation

    @v_args(inline=True)  # Children are provided as *args instead of a list argument
    def requirement_indicator(self, requirement_indicator: RequirementIndicator) -> AhbExpressionEvaluationResult:
        """
        If there is no condition expression but only a requirement indicator,
        all evaluations are automatically set to True.
        """
        return AhbExpressionEvaluationResult(
            requirement_indicator=requirement_indicator,
            requirement_constraint_evaluation_result=RequirementConstraintEvaluationResult(
                requirement_constraints_fulfilled=True,
                requirement_is_conditional=False,
                format_constraints_expression=None,
                hints=None,
            ),
            format_constraint_evaluation_result=FormatConstraintEvaluationResult(
                format_constraints_fulfilled=True, error_message=None
            ),
        )

    # pylint: disable=(line-too-long)
    def ahb_expression(
        self,
        list_of_single_requirement_indicator_expressions: List[
            Union[AhbExpressionEvaluationResult, Awaitable[AhbExpressionEvaluationResult]]
        ],
    ) -> Awaitable[AhbExpressionEvaluationResult]:
        """
        Returns the requirement indicator with its condition expressions already evaluated to booleans.
        If there are more than one modal mark the first whose conditions are fulfilled is returned or the
        last of the list if they are all unfulfilled.
        """
        return self._ahb_expression_async(list_of_single_requirement_indicator_expressions)

    async def _ahb_expression_async(
        self,
        list_of_single_requirement_indicator_expressions: List[
            Union[AhbExpressionEvaluationResult, Awaitable[AhbExpressionEvaluationResult]]
        ],
    ) -> AhbExpressionEvaluationResult:
        # the thing is that some user funcs (like e.g. 'requirement_indicator') are not async and there's no reason to
        # make them async. So here we have a list that is mixed: It contains both evaluation results and awaitable
        # evaluation results. The utility function 'gather_if_necessary' accounts for that (see its separate tests).
        results = await gather_if_necessary(list_of_single_requirement_indicator_expressions)
        for single_requirement_indicator_expression in results:
            if (
                single_requirement_indicator_expression.requirement_constraint_evaluation_result.requirement_constraints_fulfilled
            ):
                # if there are more than one modal mark, the requirement is conditional
                # even if the one of modal mark itself is not, e.g. `Muss[1] Kann`
                if len(list_of_single_requirement_indicator_expressions) > 1:
                    single_requirement_indicator_expression.requirement_constraint_evaluation_result.requirement_is_conditional = (
                        True
                    )
                return single_requirement_indicator_expression
        return results[-1]


async def evaluate_ahb_expression_tree(
    parsed_tree: Tree, entered_input: Optional[str]
) -> AhbExpressionEvaluationResult:
    """
    Evaluates the tree built from the ahb expressions with the help of the AhbExpressionTransformer.

    :param parsed_tree: Tree
    :param entered_input: the conditions as defined in the AHBs in the form of ConditionNodes
    :return: the result of the overall condition check (including requirement constraints, format constraints,
        several modal marks)
    """
    try:
        result = AhbExpressionTransformer(entered_input).transform(parsed_tree)
    except VisitError as visit_err:
        raise visit_err.orig_exc

    return await result



"""
Test the maus provider as a concept and the file based maus provider as an implementation.
"""
import json
from pathlib import Path

import pytest  # type:ignore[import]
from maus import DeepAnwendungshandbuch
from maus.edifact import EdifactFormat, EdifactFormatVersion
from maus.models.anwendungshandbuch import AhbMetaInformation, DeepAnwendungshandbuchSchema

from ahbicht.validation.maus_provider import FileBasedMausProvider, MausProvider

pytestmark = pytest.mark.asyncio


class MyFooBarMausProvider(FileBasedMausProvider):
    """
    A maus provider, just for this test.
    """

    def to_path(
        self, edifact_format: EdifactFormat, edifact_format_version: EdifactFormatVersion, pruefidentifikator: str
    ) -> Path:
        return Path(f"{edifact_format_version}/{edifact_format}/{pruefidentifikator}_maus.json")


class TestMausProvider:
    """
    Test the file based maus provider
    """

    def test_file_based_maus_provider_not_found(self, tmpdir_factory):
        maus_root_dir = tmpdir_factory.mktemp("test_dir")
        # we just create an empty temporary directory, just for the sake of having a directory
        provider: MausProvider = MyFooBarMausProvider(base_path=maus_root_dir)
        assert provider is not None
        actual = provider.get_maus(
            edifact_format=EdifactFormat.UTILMD,
            edifact_format_version=EdifactFormatVersion.FV2104,
            pruefidentifikator="11001",
        )
        assert actual is None  # because nothing was found

    def test_file_based_maus_provider_success(self, tmpdir_factory):
        maus_root_dir = tmpdir_factory.mktemp("test_dir")
        # no we create something more: a directory structure compatible with the above MyFooBarMausProvider
        maus_root_dir.mkdir("FV2104")
        maus_root_dir.mkdir("FV2104/UTILMD")
        # a minimal maus that is serializable and deserializable
        example_maus = DeepAnwendungshandbuch(meta=AhbMetaInformation(pruefidentifikator="11001"), lines=[])
        with open(maus_root_dir / "FV2104/UTILMD/11001_maus.json", "w+") as maus_test_outfile:
            deep_ahb_dict = DeepAnwendungshandbuchSchema().dump(example_maus)  # create a dictionary
            json.dump(deep_ahb_dict, maus_test_outfile)  # dump the dictionary to the file
        provider: MausProvider = MyFooBarMausProvider(base_path=maus_root_dir)
        assert provider is not None
        actual = provider.get_maus(
            edifact_format=EdifactFormat.UTILMD,
            edifact_format_version=EdifactFormatVersion.FV2104,
            pruefidentifikator="11001",
        )
        assert actual is not None  # because the file was found
        assert actual == example_maus  # is equivalent to the original because it was read from the file
